"""
Copyright (c) 2018 Noderabbit Inc., d.b.a. Appsembler

All rights reserved.

MIT License

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.

See https://nicksnettravels.builttoroam.com/post/2017/01/24/Verifying-Azure-Active-Directory-JWT-Tokens.aspx
    for verifying JWT tokens.
"""

import json
import six

from cryptography.hazmat.primitives import serialization
from jwt import DecodeError, ExpiredSignature, decode as jwt_decode
from jwt.utils import base64url_decode


try:
    from jwt.algorithms import RSAAlgorithm
except ImportError as e:
    raise Exception(
        # Python 3.3 is not supported because of compatibility in
        # Cryptography package in Python3.3 You are welcome to patch
        # and open a pull request.
        'Cryptography library is required for this backend ' \
        '(AzureADB2COAuth2) to work. Note that this backend is only ' \
        'supported on Python 2 and Python 3.4+.'
    )

from ..exceptions import AuthException, AuthTokenError
from .azuread import AzureADOAuth2


class AzureADB2COAuth2(AzureADOAuth2):
    name = 'azuread-b2c-oauth2'

    BASE_URL = 'https://login.microsoftonline.com/{tenant_id}'
    AUTHORIZATION_URL = '{base_url}/oauth2/v2.0/authorize'
    OPENID_CONFIGURATION_URL = '{base_url}/v2.0/.well-known/openid-configuration?p={policy}'
    ACCESS_TOKEN_URL = '{base_url}/oauth2/v2.0/token?p={policy}'
    JWKS_URL = '{base_url}/discovery/v2.0/keys?p={policy}'
    DEFAULT_SCOPE = ['openid', 'email']
    EXTRA_DATA = [
        ('access_token', 'access_token'),
        ('id_token', 'id_token'),
        ('refresh_token', 'refresh_token'),
        ('id_token_expires_in', 'expires'),
        ('exp', 'expires_on'),
        ('not_before', 'not_before'),
        ('given_name', 'first_name'),
        ('family_name', 'last_name'),
        ('tfp', 'policy'),
        ('token_type', 'token_type')
    ]

    @property
    def tenant_id(self):
        return self.setting('TENANT_ID', 'common')

    @property
    def policy(self):
        policy = self.setting('POLICY')
        if not policy or not policy.lower().startswith('b2c_'):
            raise AuthException('SOCIAL_AUTH_AZUREAD_B2C_OAUTH2_POLICY is '
                                'required and should start with `b2c_`')
        return policy

    @property
    def base_url(self):
        return self.BASE_URL.format(tenant_id=self.tenant_id)

    def openid_configuration_url(self):
        return self.OPENID_CONFIGURATION_URL.format(base_url=self.base_url,
                                                    policy=self.policy)

    def authorization_url(self):
        # Policy is required, but added later by `auth_extra_arguments()`
        return self.AUTHORIZATION_URL.format(base_url=self.base_url)

    def access_token_url(self):
        return self.ACCESS_TOKEN_URL.format(base_url=self.base_url,
                                            policy=self.policy)

    def jwks_url(self):
        return self.JWKS_URL.format(base_url=self.base_url,
                                    policy=self.policy)

    def request_access_token(self, *args, **kwargs):
        """
        This is probably a hack, but otherwise AzureADOAuth2 expects
        `access_token`.

        However, B2C backends provides `id_token`.
        """
        response = super(AzureADB2COAuth2, self).request_access_token(
            *args,
            **kwargs
        )
        if 'access_token' not in response:
            response['access_token'] = response['id_token']
        return response

    def auth_extra_arguments(self):
        """
        Return extra arguments needed on auth process.

        The defaults can be overridden by GET parameters.
        """
        extra_arguments = super(AzureADB2COAuth2, self).auth_extra_arguments()
        extra_arguments['p'] = self.policy or self.data.get('p')
        return extra_arguments

    def jwt_key_to_pem(self, key_json_dict):
        """
        Builds a PEM formatted key string from a JWT public key dict.
        """
        pub_key = RSAAlgorithm.from_jwk(json.dumps(key_json_dict))
        return pub_key.public_bytes(
            encoding=serialization.Encoding.PEM,
            format=serialization.PublicFormat.SubjectPublicKeyInfo
        )

    def get_user_id(self, details, response):
        """Use subject (sub) claim as unique id."""
        return response.get('sub')

    def get_user_details(self, response):
        """
        Email address is returned on a different attribute for AzureAD
        B2C backends.
        """
        details = super(AzureADB2COAuth2, self).get_user_details(response)
        if not details['email'] and response.get('emails'):
            details['email'] = response['emails'][0]
        return details

    def get_public_key(self, kid):
        """
        Retrieve JWT keys from the URL.
        """
        resp = self.request(self.jwks_url(), method='GET')
        resp.raise_for_status()

        # find the proper key for the kid
        for key in resp.json()['keys']:
            if key['kid'] == kid:
                return self.jwt_key_to_pem(key)
        raise DecodeError('Cannot find kid={}'.format(kid))

    def user_data(self, access_token, *args, **kwargs):
        response = kwargs.get('response')

        id_token = response.get('id_token')
        if six.PY2:
            # str() to fix a bug in Python's base64
            # https://stackoverflow.com/a/2230623/161278
            id_token = str(id_token)

        jwt_header_json = base64url_decode(id_token.split('.')[0])
        jwt_header = json.loads(jwt_header_json.decode('ascii'))

        # `kid` is short for key id
        key = self.get_public_key(jwt_header['kid'])

        try:
            return jwt_decode(
                id_token,
                key=key,
                algorithms=jwt_header['alg'],
                audience=self.setting('KEY'),
                leeway=self.setting('JWT_LEEWAY', default=0),
            )
        except (DecodeError, ExpiredSignature) as error:
            raise AuthTokenError(self, error)



from .oauth import BaseOAuth2


class ChangeTipOAuth2(BaseOAuth2):
    """ChangeTip OAuth authentication backend
       https://www.changetip.com/api
    """
    name = 'changetip'
    AUTHORIZATION_URL = 'https://www.changetip.com/o/authorize/'
    ACCESS_TOKEN_URL = 'https://www.changetip.com/o/token/'
    ACCESS_TOKEN_METHOD = 'POST'
    SCOPE_SEPARATOR = ' '

    def get_user_details(self, response):
        """Return user details from ChangeTip account"""
        return {
            'username': response['username'],
            'email': response.get('email', ''),
            'first_name': '',
            'last_name': '',
        }

    def user_data(self, access_token, *args, **kwargs):
        """Loads user data from service"""
        return self.get_json('https://api.changetip.com/v2/me/', params={
            'access_token': access_token
        })



"""
Edmodo OAuth2 Sign-in backend, docs at:
    https://python-social-auth.readthedocs.io/en/latest/backends/edmodo.html
"""
from .oauth import BaseOAuth2


class EdmodoOAuth2(BaseOAuth2):
    """Edmodo OAuth2"""
    name = 'edmodo'
    AUTHORIZATION_URL = 'https://api.edmodo.com/oauth/authorize'
    ACCESS_TOKEN_URL = 'https://api.edmodo.com/oauth/token'
    ACCESS_TOKEN_METHOD = 'POST'

    def get_user_details(self, response):
        """Return user details from Edmodo account"""
        fullname, first_name, last_name = self.get_user_names(
            first_name=response.get('first_name'),
            last_name=response.get('last_name')
        )
        return {
            'username': response.get('username'),
            'email': response.get('email'),
            'fullname': fullname,
            'first_name': first_name,
            'last_name': last_name
        }

    def user_data(self, access_token, *args, **kwargs):
        """Loads user data from Edmodo"""
        return self.get_json(
            'https://api.edmodo.com/users/me',
            params={'access_token': access_token}
        )



"""
AOL OpenId backend, docs at:
    https://python-social-auth.readthedocs.io/en/latest/backends/aol.html
"""
from .open_id import OpenIdAuth


class AOLOpenId(OpenIdAuth):
    name = 'aol'
    URL = 'http://openid.aol.com'



"""
Phabricator OAuth2 backend, docs at:
    https://secure.phabricator.com/book/phabcontrib/article/using_oauthserver/
"""
from .oauth import BaseOAuth2


class PhabricatorOAuth2(BaseOAuth2):
    """Phabricator OAuth authentication backend"""
    name = 'phabricator'
    API_URL = 'https://secure.phabricator.com'
    AUTHORIZATION_URL = 'https://secure.phabricator.com/oauthserver/auth/'
    ACCESS_TOKEN_URL = 'https://secure.phabricator.com/oauthserver/token/'
    ACCESS_TOKEN_METHOD = 'POST'
    REDIRECT_STATE = False

    def api_url(self, path):
        api_url = self.setting('API_URL') or self.API_URL
        return '{0}{1}'.format(api_url.rstrip('/'), path)

    def authorization_url(self):
        return self.api_url('/oauthserver/auth/')

    def access_token_url(self):
        return self.api_url('/oauthserver/token/')

    def get_user_details(self, response):
        """Return user details from Phabricator"""
        fullname, first_name, last_name = self.get_user_names(
            response.get('realName')
        )

        return {
            'id': response.get('phid'),
            'username': response.get('userName'),
            'email': response.get('primaryEmail', ''),
            'fullname': fullname,
            'first_name': first_name,
            'last_name': last_name,
        }

    def user_data(self, access_token, *args, **kwargs):
        """Loads user data from API"""
        return self.get_json(self.api_url('/api/user.whoami'), params={
            'access_token': access_token,
        })



import json

from .oauth import OAuth2Test


class DisqusOAuth2Test(OAuth2Test):
    backend_path = 'social_core.backends.disqus.DisqusOAuth2'
    user_data_url = 'https://disqus.com/api/3.0/users/details.json'
    expected_username = 'foobar'
    access_token_body = json.dumps({
        'access_token': 'foobar',
        'token_type': 'bearer'
    })
    user_data_body = json.dumps({
        'code': 0,
        'response': {
            'username': 'foobar',
            'numFollowers': 0,
            'isFollowing': False,
            'numFollowing': 0,
            'name': 'Foo Bar',
            'numPosts': 0,
            'url': '',
            'isAnonymous': False,
            'rep': 1.231755,
            'about': '',
            'isFollowedBy': False,
            'connections': {},
            'emailHash': '5280f14cedf530b544aecc31fcfe0240',
            'reputation': 1.231755,
            'avatar': {
                'small': {
                    'permalink': 'https://disqus.com/api/users/avatars/'
                                 'foobar.jpg',
                    'cache': 'https://securecdn.disqus.com/uploads/'
                             'users/453/4556/avatar32.jpg?1285535379'
                },
                'isCustom': False,
                'permalink': 'https://disqus.com/api/users/avatars/foobar.jpg',
                'cache': 'https://securecdn.disqus.com/uploads/users/453/'
                         '4556/avatar92.jpg?1285535379',
                'large': {
                    'permalink': 'https://disqus.com/api/users/avatars/'
                                 'foobar.jpg',
                    'cache': 'https://securecdn.disqus.com/uploads/users/'
                             '453/4556/avatar92.jpg?1285535379'
                }
            },
            'profileUrl': 'http://disqus.com/foobar/',
            'numLikesReceived': 0,
            'isPrimary': True,
            'joinedAt': '2010-09-26T21:09:39',
            'id': '1010101',
            'location': ''
        }
    })

    def test_login(self):
        self.do_login()

    def test_partial_pipeline(self):
        self.do_partial_pipeline()



"""
Yahoo OpenId, OAuth1 and OAuth2 backends, docs at:
    https://python-social-auth.readthedocs.io/en/latest/backends/yahoo.html
"""
from requests.auth import HTTPBasicAuth

from ..utils import handle_http_errors
from .open_id import OpenIdAuth
from .oauth import BaseOAuth2, BaseOAuth1


class YahooOpenId(OpenIdAuth):
    """Yahoo OpenID authentication backend"""
    name = 'yahoo'
    URL = 'http://me.yahoo.com'


class YahooOAuth(BaseOAuth1):
    """Yahoo OAuth authentication backend. DEPRECATED"""
    name = 'yahoo-oauth'
    ID_KEY = 'guid'
    AUTHORIZATION_URL = 'https://api.login.yahoo.com/oauth/v2/request_auth'
    REQUEST_TOKEN_URL = \
        'https://api.login.yahoo.com/oauth/v2/get_request_token'
    ACCESS_TOKEN_URL = 'https://api.login.yahoo.com/oauth/v2/get_token'
    EXTRA_DATA = [
        ('guid', 'id'),
        ('access_token', 'access_token'),
        ('expires', 'expires')
    ]

    def get_user_details(self, response):
        """Return user details from Yahoo Profile"""
        fullname, first_name, last_name = self.get_user_names(
            first_name=response.get('givenName'),
            last_name=response.get('familyName')
        )
        emails = [email for email in response.get('emails', [])
                        if email.get('handle')]
        emails.sort(key=lambda e: e.get('primary', False), reverse=True)
        return {'username': response.get('nickname'),
                'email': emails[0]['handle'] if emails else '',
                'fullname': fullname,
                'first_name': first_name,
                'last_name': last_name}

    def user_data(self, access_token, *args, **kwargs):
        """Loads user data from service"""
        url = 'https://social.yahooapis.com/v1/user/{0}/profile?format=json'
        return self.get_json(
            url.format(self._get_guid(access_token)),
            auth=self.oauth_auth(access_token)
        )['profile']

    def _get_guid(self, access_token):
        """
        Beause you have to provide GUID for every API request it's also
        returned during one of OAuth calls
        """
        return self.get_json(
            'https://social.yahooapis.com/v1/me/guid?format=json',
            auth=self.oauth_auth(access_token)
        )['guid']['value']


class YahooOAuth2(BaseOAuth2):
    """Yahoo OAuth2 authentication backend"""
    name = 'yahoo-oauth2'
    ID_KEY = 'guid'
    AUTHORIZATION_URL = 'https://api.login.yahoo.com/oauth2/request_auth'
    ACCESS_TOKEN_URL = 'https://api.login.yahoo.com/oauth2/get_token'
    ACCESS_TOKEN_METHOD = 'POST'
    EXTRA_DATA = [
        ('xoauth_yahoo_guid', 'id'),
        ('access_token', 'access_token'),
        ('expires_in', 'expires'),
        ('refresh_token', 'refresh_token'),
        ('token_type', 'token_type'),
    ]

    def get_user_names(self, first_name, last_name):
        if first_name or last_name:
            return ' '.join((first_name, last_name)), first_name, last_name
        return None, None, None

    def get_user_details(self, response):
        """
        Return user details from Yahoo Profile.
        To Get user email you need the profile private read permission.
        """
        fullname, first_name, last_name = self.get_user_names(
            first_name=response.get('givenName'),
            last_name=response.get('familyName')
        )
        emails = [email for email in response.get('emails', [])
                        if 'handle' in email]
        emails.sort(key=lambda e: e.get('primary', False), reverse=True)
        email = emails[0]['handle'] if emails else response.get('guid', '')
        return {
            'username': response.get('nickname'),
            'email': email,
            'fullname': fullname,
            'first_name': first_name,
            'last_name': last_name
        }

    def user_data(self, access_token, *args, **kwargs):
        """Loads user data from service"""
        url = 'https://social.yahooapis.com/v1/user/{0}/profile?format=json' \
                .format(kwargs['response']['xoauth_yahoo_guid'])
        return self.get_json(url, headers={
            'Authorization': 'Bearer {0}'.format(access_token)
        }, method='GET')['profile']

    @handle_http_errors
    def auth_complete(self, *args, **kwargs):
        """Completes loging process, must return user instance"""
        self.process_error(self.data)
        response = self.request_access_token(
            self.ACCESS_TOKEN_URL,
            auth=HTTPBasicAuth(*self.get_key_and_secret()),
            data=self.auth_complete_params(self.validate_state()),
            headers=self.auth_headers(),
            method=self.ACCESS_TOKEN_METHOD
        )
        self.process_error(response)
        return self.do_auth(response['access_token'], response=response,
                            *args, **kwargs)

    def refresh_token_params(self, token, *args, **kwargs):
        return {
            'refresh_token': token,
            'grant_type': 'refresh_token',
            'redirect_uri': 'oob',  # out of bounds
        }

    def refresh_token(self, token, *args, **kwargs):
        params = self.refresh_token_params(token, *args, **kwargs)
        url = self.REFRESH_TOKEN_URL or self.ACCESS_TOKEN_URL
        method = self.REFRESH_TOKEN_METHOD
        key = 'params' if method == 'GET' else 'data'
        request_args = {
            'headers': self.auth_headers(),
            'method': method,
            key: params
        }
        request = self.request(
            url,
            auth=HTTPBasicAuth(*self.get_key_and_secret()),
            **request_args
        )
        return self.process_refresh_token_response(request, *args, **kwargs)

    def auth_complete_params(self, state=None):
        return {
            'grant_type': 'authorization_code',  # request auth code
            'code': self.data.get('code', ''),  # server response code
            'redirect_uri': self.get_redirect_uri(state)
        }



"""
Legacy Email backend, docs at:
    https://python-social-auth.readthedocs.io/en/latest/backends/email.html
"""
from .legacy import LegacyAuth


class EmailAuth(LegacyAuth):
    name = 'email'
    ID_KEY = 'email'
    REQUIRES_EMAIL_VALIDATION = True
    EXTRA_DATA = ['email']



"""
Belgium EID OpenId backend, docs at:
    https://python-social-auth.readthedocs.io/en/latest/backends/belgium_eid.html
"""
from .open_id import OpenIdAuth


class BelgiumEIDOpenId(OpenIdAuth):
    """Belgium e-ID OpenID authentication backend"""
    name = 'belgiumeid'
    URL = 'https://www.e-contract.be/eid-idp/endpoints/openid/auth'



from .goclio import GoClioOAuth2


class GoClioEuOAuth2(GoClioOAuth2):
    name = 'goclioeu'
    AUTHORIZATION_URL = 'https://app.goclio.eu/oauth/authorize/'
    ACCESS_TOKEN_URL = 'https://app.goclio.eu/oauth/token/'

    def user_data(self, access_token, *args, **kwargs):
        """Loads user data from service"""
        return self.get_json(
            'https://app.goclio.eu/api/v2/users/who_am_i',
            params={'access_token': access_token}
        )



class SocialAuthBaseException(ValueError):
    """Base class for pipeline exceptions."""
    pass


class WrongBackend(SocialAuthBaseException):
    def __init__(self, backend_name):
        self.backend_name = backend_name

    def __str__(self):
        return 'Incorrect authentication service "{0}"'.format(
            self.backend_name
        )


class MissingBackend(WrongBackend):
    def __str__(self):
        return 'Missing backend "{0}" entry'.format(self.backend_name)


class NotAllowedToDisconnect(SocialAuthBaseException):
    """User is not allowed to disconnect it's social account."""
    pass


class AuthException(SocialAuthBaseException):
    """Auth process exception."""
    def __init__(self, backend, *args, **kwargs):
        self.backend = backend
        super(AuthException, self).__init__(*args, **kwargs)


class AuthFailed(AuthException):
    """Auth process failed for some reason."""
    def __str__(self):
        msg = super(AuthFailed, self).__str__()
        if msg == 'access_denied':
            return 'Authentication process was canceled'
        return 'Authentication failed: {0}'.format(msg)


class AuthCanceled(AuthException):
    """Auth process was canceled by user."""
    def __init__(self, *args, **kwargs):
        self.response = kwargs.pop('response', None)
        super(AuthCanceled, self).__init__(*args, **kwargs)

    def __str__(self):
        msg = super(AuthCanceled, self).__str__()
        if msg:
            return 'Authentication process canceled: {0}'.format(msg)
        return 'Authentication process canceled'


class AuthUnknownError(AuthException):
    """Unknown auth process error."""
    def __str__(self):
        msg = super(AuthUnknownError, self).__str__()
        return 'An unknown error happened while authenticating {0}'.format(msg)


class AuthTokenError(AuthException):
    """Auth token error."""
    def __str__(self):
        msg = super(AuthTokenError, self).__str__()
        return 'Token error: {0}'.format(msg)


class AuthMissingParameter(AuthException):
    """Missing parameter needed to start or complete the process."""
    def __init__(self, backend, parameter, *args, **kwargs):
        self.parameter = parameter
        super(AuthMissingParameter, self).__init__(backend, *args, **kwargs)

    def __str__(self):
        return 'Missing needed parameter {0}'.format(self.parameter)


class AuthStateMissing(AuthException):
    """State parameter is incorrect."""
    def __str__(self):
        return 'Session value state missing.'


class AuthStateForbidden(AuthException):
    """State parameter is incorrect."""
    def __str__(self):
        return 'Wrong state parameter given.'


class AuthAlreadyAssociated(AuthException):
    """A different user has already associated the target social account"""
    pass


class AuthTokenRevoked(AuthException):
    """User revoked the access_token in the provider."""
    def __str__(self):
        return 'User revoke access to the token'


class AuthForbidden(AuthException):
    """Authentication for this user is forbidden"""
    def __str__(self):
        return 'Your credentials aren\'t allowed'


class AuthUnreachableProvider(AuthException):
    """Cannot reach the provider"""
    def __str__(self):
        return 'The authentication provider could not be reached'


class InvalidEmail(AuthException):
    def __str__(self):
        return 'Email couldn\'t be validated'



"""
Appsfueld OAuth2 backend (with sandbox mode support), docs at:
    https://python-social-auth.readthedocs.io/en/latest/backends/appsfuel.html
"""
from .oauth import BaseOAuth2


class AppsfuelOAuth2(BaseOAuth2):
    name = 'appsfuel'
    ID_KEY = 'user_id'
    AUTHORIZATION_URL = 'http://app.appsfuel.com/content/permission'
    ACCESS_TOKEN_URL = 'https://api.appsfuel.com/v1/live/oauth/token'
    ACCESS_TOKEN_METHOD = 'POST'
    USER_DETAILS_URL = 'https://api.appsfuel.com/v1/live/user'

    def get_user_details(self, response):
        """Return user details from Appsfuel account"""
        email = response.get('email', '')
        username = email.split('@')[0] if email else ''
        fullname, first_name, last_name = self.get_user_names(
            response.get('display_name', '')
        )
        return {
            'username': username,
            'fullname': fullname,
            'first_name': first_name,
            'last_name': last_name,
            'email': email
        }

    def user_data(self, access_token, *args, **kwargs):
        """Loads user data from service"""
        return self.get_json(self.USER_DETAILS_URL, params={
            'access_token': access_token
        })


class AppsfuelOAuth2Sandbox(AppsfuelOAuth2):
    name = 'appsfuel-sandbox'
    AUTHORIZATION_URL = 'https://api.appsfuel.com/v1/sandbox/choose'
    ACCESS_TOKEN_URL = 'https://api.appsfuel.com/v1/sandbox/oauth/token'
    USER_DETAILS_URL = 'https://api.appsfuel.com/v1/sandbox/user'



"""
Vend  OAuth2 backend:
"""
from .oauth import BaseOAuth2


class VendOAuth2(BaseOAuth2):
    name = 'vend'
    AUTHORIZATION_URL = 'https://secure.vendhq.com/connect'
    ACCESS_TOKEN_URL = 'https://{0}.vendhq.com/api/1.0/token'
    ACCESS_TOKEN_METHOD = 'POST'
    REDIRECT_STATE = False
    EXTRA_DATA = [
        ('refresh_token', 'refresh_token'),
        ('domain_prefix', 'domain_prefix')
    ]

    def access_token_url(self):
        return self.ACCESS_TOKEN_URL.format(self.data['domain_prefix'])

    def get_user_details(self, response):
        email = response['email']
        username = response.get('username') or email.split('@', 1)[0]
        return {
            'username': username,
            'email': email,
            'fullname': '',
            'first_name': '',
            'last_name': ''
        }

    def user_data(self, access_token, *args, **kwargs):
        """Loads user data from service"""
        prefix = kwargs['response']['domain_prefix']
        url = 'https://{0}.vendhq.com/api/users'.format(prefix)
        data = self.get_json(url, headers={
            'Authorization': 'Bearer {0}'.format(access_token)
        })
        return data['users'][0] if data.get('users') else {}



import json

from .oauth import OAuth2Test


class StripeOAuth2Test(OAuth2Test):
    backend_path = 'social_core.backends.stripe.StripeOAuth2'
    access_token_body = json.dumps({
        'stripe_publishable_key': 'pk_test_foobar',
        'access_token': 'foobar',
        'livemode': False,
        'token_type': 'bearer',
        'scope': 'read_only',
        'refresh_token': 'rt_foobar',
        'stripe_user_id': 'acct_foobar'
    })
    expected_username = 'acct_foobar'

    def test_login(self):
        self.do_login()

    def test_partial_pipeline(self):
        self.do_partial_pipeline()



from functools import wraps

from .utils import partial_prepare


def partial_step(save_to_session):
    """Wraps func to behave like a partial pipeline step, any output
    that's not None or {} will be considered a response object and
    will be returned to user.

    The pipeline function will receive a current_partial object, it
    contains the partial pipeline data and a token that is used to
    identify it when it's continued, this is useful to build links
    with the token.

    The default value for this parameter is partial_token, but can be
    overridden by SOCIAL_AUTH_PARTIAL_PIPELINE_TOKEN_NAME setting.

    The token is also stored in the session under the
    partial_pipeline_token key when the save_to_session parameter is True.
    """
    def decorator(func):
        @wraps(func)
        def wrapper(strategy, backend, pipeline_index, *args, **kwargs):
            current_partial = partial_prepare(strategy, backend, pipeline_index,
                                              *args, **kwargs)

            out = func(strategy=strategy,
                       backend=backend,
                       pipeline_index=pipeline_index,
                       current_partial=current_partial,
                       *args, **kwargs) or {}

            if not isinstance(out, dict):
                strategy.storage.partial.store(current_partial)
                if save_to_session:
                    strategy.session_set('partial_pipeline_token', current_partial.token)
            return out
        return wrapper
    return decorator


# Backward compatible partial decorator, that stores the token in the session
partial = partial_step(save_to_session=True)



import json

from .oauth import OAuth2Test


class CourseraOAuth2Test(OAuth2Test):
    backend_path = 'social_core.backends.coursera.CourseraOAuth2'
    user_data_url = \
        'https://api.coursera.org/api/externalBasicProfiles.v1?q=me'
    expected_username = '560e7ed2076e0d589e88bd74b6aad4b7'
    access_token_body = json.dumps({
        'access_token': 'foobar',
        'token_type': 'Bearer',
        'expires_in': 1795
    })
    request_token_body = json.dumps({
        'code': 'foobar-code',
        'client_id': 'foobar-client-id',
        'client_secret': 'foobar-client-secret',
        'redirect_uri': 'http://localhost:8000/accounts/coursera/',
        'grant_type': 'authorization_code'
    })
    user_data_body = json.dumps({
        'token_type': 'Bearer',
        'paging': None,
        'elements': [{
            'id': '560e7ed2076e0d589e88bd74b6aad4b7'
        }],
        'access_token': 'foobar',
        'expires_in': 1800,
        'linked': None
    })

    def test_login(self):
        self.do_login()

    def test_partial_pipeline(self):
        self.do_partial_pipeline()



"""
Qiita OAuth2 backend, docs at:
    https://python-social-auth.readthedocs.io/en/latest/backends/qiita.html
    http://qiita.com/api/v2/docs#get-apiv2oauthauthorize
"""
import json

from .oauth import BaseOAuth2


class QiitaOAuth2(BaseOAuth2):
    """Qiita OAuth authentication backend"""
    name = 'qiita'

    AUTHORIZATION_URL = 'https://qiita.com/api/v2/oauth/authorize'
    ACCESS_TOKEN_URL = 'https://qiita.com/api/v2/access_tokens'
    ACCESS_TOKEN_METHOD = 'POST'
    SCOPE_SEPARATOR = ' '
    REDIRECT_STATE = True
    EXTRA_DATA = [
        ('description', 'description'),
        ('facebook_id', 'facebook_id'),
        ('followees_count', 'followees_count'),
        ('followers_count', 'followers_count'),
        ('github_login_name', 'github_login_name'),
        ('id', 'id'),
        ('items_count', 'items_count'),
        ('linkedin_id', 'linkedin_id'),
        ('location', 'location'),
        ('name', 'name'),
        ('organization', 'organization'),
        ('profile_image_url', 'profile_image_url'),
        ('twitter_screen_name', 'twitter_screen_name'),
        ('website_url', 'website_url'),
    ]

    def auth_complete_params(self, state=None):
        data = super(QiitaOAuth2, self).auth_complete_params(state)
        if "grant_type" in data:
            del data["grant_type"]
        if "redirect_uri" in data:
            del data["redirect_uri"]
        return json.dumps(data)

    def auth_headers(self):
        return {'Content-Type': 'application/json'}

    def request_access_token(self, *args, **kwargs):
        data = super(QiitaOAuth2, self).request_access_token(*args, **kwargs)
        data.update({'access_token': data['token']})
        return data

    def get_user_details(self, response):
        """Return user details from Qiita account"""
        return {
            'username': response['id'],
            'fullname': response['name'],
        }

    def user_data(self, access_token, *args, **kwargs):
        """Loads user data from service"""
        return self.get_json(
            'https://qiita.com/api/v2/authenticated_user',
           headers={
                'Authorization': 'Bearer {0}'.format(access_token)
            })



import json

from .oauth import OAuth2Test


class StravaOAuthTest(OAuth2Test):
    backend_path = 'social_core.backends.strava.StravaOAuth'
    user_data_url = 'https://www.strava.com/api/v3/athlete'
    expected_username = '227615'
    access_token_body = json.dumps({
      "access_token": "83ebeabdec09f6670863766f792ead24d61fe3f9",
      "athlete": {
        "id": 227615,
        "resource_state": 3,
        "firstname": "John",
        "lastname": "Applestrava",
        "profile_medium": "http://pics.com/227615/medium.jpg",
        "profile": "http://pics.com/227615/large.jpg",
        "city": "San Francisco",
        "state": "California",
        "country": "United States",
        "sex": "M",
        "friend": "null",
        "follower": "null",
        "premium": "true",
        "created_at": "2008-01-01T17:44:00Z",
        "updated_at": "2013-09-04T20:00:50Z",
        "follower_count": 273,
        "friend_count": 19,
        "mutual_friend_count": 0,
        "date_preference": "%m/%d/%Y",
        "measurement_preference": "feet",
        "email": "john@applestrava.com",
        "clubs": [],
        "bikes": [],
        "shoes": []
      }
    })
    user_data_body = json.dumps({
      "id": 227615,
      "resource_state": 2,
      "firstname": "John",
      "lastname": "Applestrava",
      "profile_medium": "http://pics.com/227615/medium.jpg",
      "profile": "http://pics.com/227615/large.jpg",
      "city": "San Francisco",
      "state": "CA",
      "country": "United States",
      "sex": "M",
      "friend": "null",
      "follower": "accepted",
      "premium": "true",
      "created_at": "2011-03-19T21:59:57Z",
      "updated_at": "2013-09-05T16:46:54Z",
      "approve_followers": "false"
    })

    def test_login(self):
        self.do_login()

    def test_partial_pipeline(self):
        self.do_partial_pipeline()



import json

from .oauth import OAuth2Test


class YammerOAuth2Test(OAuth2Test):
    backend_path = 'social_core.backends.yammer.YammerOAuth2'
    expected_username = 'foobar'
    access_token_body = json.dumps({
        'access_token': {
            'user_id': 1010101010,
            'view_groups': True,
            'modify_messages': True,
            'network_id': 101010,
            'created_at': '2013/03/17 16:39:56 +0000',
            'view_members': True,
            'authorized_at': '2013/03/17 16:39:56 +0000',
            'view_subscriptions': True,
            'view_messages': True,
            'modify_subscriptions': True,
            'token': 'foobar',
            'expires_at': None,
            'network_permalink': 'foobar.com',
            'view_tags': True,
            'network_name': 'foobar.com'
        },
        'user': {
            'last_name': 'Bar',
            'web_url': 'https://www.yammer.com/foobar/users/foobar',
            'expertise': None,
            'full_name': 'Foo Bar',
            'timezone': 'Pacific Time (US & Canada)',
            'mugshot_url': 'https://mug0.assets-yammer.com/mugshot/images/'
                           '48x48/no_photo.png',
            'guid': None,
            'network_name': 'foobar',
            'id': 1010101010,
            'previous_companies': [],
            'first_name': 'Foo',
            'stats': {
                'following': 0,
                'followers': 0,
                'updates': 1
            },
            'hire_date': None,
            'state': 'active',
            'location': None,
            'department': 'Software Development',
            'type': 'user',
            'show_ask_for_photo': True,
            'job_title': 'Software Developer',
            'interests': None,
            'kids_names': None,
            'activated_at': '2013/03/17 16:27:50 +0000',
            'verified_admin': 'false',
            'can_broadcast': 'false',
            'schools': [],
            'admin': 'false',
            'network_domains': ['foobar.com'],
            'name': 'foobar',
            'external_urls': [],
            'url': 'https://www.yammer.com/api/v1/users/1010101010',
            'settings': {
                'xdr_proxy': 'https://xdrproxy.yammer.com'
            },
            'summary': None,
            'network_id': 101010,
            'contact': {
                'phone_numbers': [],
                'im': {
                    'username': '',
                    'provider': ''
                },
                'email_addresses': [{
                    'type': 'primary',
                    'address': 'foo@bar.com'
                }],
                'has_fake_email': False
            },
            'birth_date': '',
            'mugshot_url_template': 'https://mug0.assets-yammer.com/mugshot/'
                                    'images/{width}x{height}/no_photo.png',
            'significant_other': None
        },
        'network': {
            'show_upgrade_banner': False,
            'header_text_color': '#FFFFFF',
            'is_org_chart_enabled': True,
            'name': 'foobar.com',
            'is_group_enabled': True,
            'header_background_color': '#396B9A',
            'created_at': '2012/12/26 16:52:35 +0000',
            'profile_fields_config': {
                'enable_work_phone': True,
                'enable_mobile_phone': True,
                'enable_job_title': True
            },
            'permalink': 'foobar.com',
            'paid': False,
            'id': 101010,
            'is_chat_enabled': True,
            'web_url': 'https://www.yammer.com/foobar.com',
            'moderated': False,
            'community': False,
            'type': 'network',
            'navigation_background_color': '#38699F',
            'navigation_text_color': '#FFFFFF'
        }
    })

    def test_login(self):
        self.do_login()

    def test_partial_pipeline(self):
        self.do_partial_pipeline()



"""
Angel OAuth2 backend, docs at:
    https://python-social-auth.readthedocs.io/en/latest/backends/angel.html
"""
from .oauth import BaseOAuth2


class AngelOAuth2(BaseOAuth2):
    name = 'angel'
    AUTHORIZATION_URL = 'https://angel.co/api/oauth/authorize/'
    ACCESS_TOKEN_METHOD = 'POST'
    ACCESS_TOKEN_URL = 'https://angel.co/api/oauth/token/'
    REDIRECT_STATE = False

    def get_user_details(self, response):
        """Return user details from Angel account"""
        username = response['angellist_url'].split('/')[-1]
        email = response.get('email', '')
        fullname, first_name, last_name = self.get_user_names(response['name'])
        return {'username': username,
                'fullname': fullname,
                'first_name': first_name,
                'last_name': last_name,
                'email': email}

    def user_data(self, access_token, *args, **kwargs):
        """Loads user data from service"""
        return self.get_json('https://api.angel.co/1/me/', params={
            'access_token': access_token
        })



import unittest2 as unittest
try:
    from mock import Mock, patch
except ImportError:
    from unittest.mock import Mock, patch

from ..pipeline.partial import partial, partial_step

class PartialDecoratorTestCase(unittest.TestCase):

    def setUp(self):
        super(PartialDecoratorTestCase, self).setUp()
        self.mock_current_partial_token = Mock()
        self.mock_current_partial = Mock(token=self.mock_current_partial_token)

        self.mock_strategy = Mock()
        self.mock_backend = Mock()
        self.mock_pipeline_index = Mock()
        self.mock_partial_store = Mock()
        self.mock_strategy.storage.partial.store = self.mock_partial_store

        self.mock_sesstion_set = Mock()
        self.mock_strategy.session_set = self.mock_sesstion_set

    def test_save_to_session(self):
        # GIVEN
        expected_response = Mock()

        @partial_step(save_to_session=True)
        def decorated_func(*args, **kwargs):
            return expected_response

        # WHEN
        with patch('social_core.pipeline.partial.partial_prepare', return_value=self.mock_current_partial):
            response = decorated_func(self.mock_strategy, self.mock_backend, self.mock_pipeline_index)

            # THEN
            self.assertEqual(expected_response, response)

            self.assertEqual(1, self.mock_partial_store.call_count)
            self.assertEqual((self.mock_current_partial,), self.mock_partial_store.call_args[0])

            self.assertEqual(1, self.mock_sesstion_set.call_count)
            self.assertEqual(('partial_pipeline_token', self.mock_current_partial_token), self.mock_sesstion_set.call_args[0])

    def test_not_to_save_to_session(self):
        # GIVEN
        expected_response = Mock()

        @partial_step(save_to_session=False)
        def decorated_func(*args, **kwargs):
            return expected_response

        # WHEN
        with patch('social_core.pipeline.partial.partial_prepare', return_value=self.mock_current_partial):
            response = decorated_func(self.mock_strategy, self.mock_backend, self.mock_pipeline_index)

            # THEN
            self.assertEqual(expected_response, response)

            self.assertEqual(1, self.mock_partial_store.call_count)
            self.assertEqual((self.mock_current_partial,), self.mock_partial_store.call_args[0])

            self.assertEqual(0, self.mock_sesstion_set.call_count)

    def test_save_to_session_by_backward_compatible_decorator(self):
        # GIVEN
        expected_response = Mock()

        @partial
        def decorated_func(*args, **kwargs):
            return expected_response

        # WHEN
        with patch('social_core.pipeline.partial.partial_prepare', return_value=self.mock_current_partial):
            response = decorated_func(self.mock_strategy, self.mock_backend, self.mock_pipeline_index)

            # THEN
            self.assertEqual(expected_response, response)

            self.assertEqual(1, self.mock_partial_store.call_count)
            self.assertEqual((self.mock_current_partial,), self.mock_partial_store.call_args[0])

            self.assertEqual(1, self.mock_sesstion_set.call_count)
            self.assertEqual(('partial_pipeline_token', self.mock_current_partial_token), self.mock_sesstion_set.call_args[0])

    def test_not_to_save_to_session_when_the_response_is_a_dict(self):
        # GIVEN
        expected_response = {'test_key': 'test_value'}

        @partial_step(save_to_session=True)
        def decorated_func(*args, **kwargs):
            return expected_response

        # WHEN
        response = decorated_func(self.mock_strategy, self.mock_backend, self.mock_pipeline_index)

        # THEN
        self.assertEqual(expected_response, response)
        self.assertEqual(0, self.mock_partial_store.call_count)
        self.assertEqual(0, self.mock_sesstion_set.call_count)



import json

from .oauth import OAuth2Test


class DigitalOceanOAuthTest(OAuth2Test):
    backend_path = 'social_core.backends.digitalocean.DigitalOceanOAuth'
    user_data_url = 'https://api.digitalocean.com/v2/account'
    expected_username = 'sammy@digitalocean.com'
    access_token_body = json.dumps({
        'access_token': '547cac21118ae7',
        'token_type': 'bearer',
        'expires_in': 2592000,
        'refresh_token': '00a3aae641658d',
        'scope': 'read write',
        'info': {
            'name': 'Sammy Shark',
            'email': 'sammy@digitalocean.com'
        }
    })
    user_data_body = json.dumps({
        "account": {
            'droplet_limit': 25,
            'email': 'sammy@digitalocean.com',
            'uuid': 'b6fr89dbf6d9156cace5f3c78dc9851d957381ef',
            'email_verified': True
        }
    })

    def test_login(self):
        self.do_login()

    def test_partial_pipeline(self):
        self.do_partial_pipeline()



import time

from .oauth import BaseOAuth2
from ..utils import handle_http_errors


class ItembaseOAuth2(BaseOAuth2):
    name = 'itembase'
    ID_KEY = 'uuid'
    AUTHORIZATION_URL = 'https://accounts.itembase.com/oauth/v2/auth'
    ACCESS_TOKEN_URL = 'https://accounts.itembase.com/oauth/v2/token'
    USER_DETAILS_URL = 'https://users.itembase.com/v1/me'
    ACTIVATION_ENDPOINT = 'https://solutionservice.itembase.com/activate'
    DEFAULT_SCOPE = ['user.minimal']
    EXTRA_DATA = [
        ('access_token', 'access_token'),
        ('token_type', 'token_type'),
        ('refresh_token', 'refresh_token'),
        ('expires_in', 'expires_in'),  # seconds to expiration
        ('expires', 'expires'),  # expiration timestamp in UTC
        ('uuid', 'uuid'),
        ('username', 'username'),
        ('email', 'email'),
        ('first_name', 'first_name'),
        ('middle_name', 'middle_name'),
        ('last_name', 'last_name'),
        ('name_format', 'name_format'),
        ('locale', 'locale'),
        ('preferred_currency', 'preferred_currency'),
    ]

    def add_expires(self, data):
        data['expires'] = int(time.time()) + data.get('expires_in', 0)
        return data

    def extra_data(self, user, uid, response, details=None, *args, **kwargs):
        data = BaseOAuth2.extra_data(self, user, uid, response,
                                     details=details,
                                     *args, **kwargs)
        return self.add_expires(data)

    def process_refresh_token_response(self, response, *args, **kwargs):
        data = BaseOAuth2.process_refresh_token_response(self, response,
                                                         *args, **kwargs)
        return self.add_expires(data)

    def get_user_details(self, response):
        """Return user details from Itembase account"""
        return response

    def user_data(self, access_token, *args, **kwargs):
        return self.get_json(self.USER_DETAILS_URL, headers={
            'Authorization': 'Bearer {0}'.format(access_token)
        })

    def activation_data(self, response):
        # returns activation_data dict with activation_url inside
        # see http://developers.itembase.com/authentication/activation
        return self.get_json(self.ACTIVATION_ENDPOINT, headers={
            'Authorization': 'Bearer {0}'.format(response['access_token'])
        })

    @handle_http_errors
    def auth_complete(self, *args, **kwargs):
        """Completes login process, must return user instance"""
        state = self.validate_state()
        self.process_error(self.data)
        # itembase needs GET request with params instead of just data
        response = self.request_access_token(
            self.access_token_url(),
            params=self.auth_complete_params(state),
            headers=self.auth_headers(),
            auth=self.auth_complete_credentials(),
            method=self.ACCESS_TOKEN_METHOD
        )
        self.process_error(response)
        return self.do_auth(response['access_token'], response=response,
                            *args, **kwargs)


class ItembaseOAuth2Sandbox(ItembaseOAuth2):
    name = 'itembase-sandbox'
    AUTHORIZATION_URL = 'http://sandbox.accounts.itembase.io/oauth/v2/auth'
    ACCESS_TOKEN_URL = 'http://sandbox.accounts.itembase.io/oauth/v2/token'
    USER_DETAILS_URL = 'http://sandbox.users.itembase.io/v1/me'
    ACTIVATION_ENDPOINT = 'http://sandbox.solutionservice.itembase.io/activate'



import time

from requests import request, ConnectionError

from ..utils import SSLHttpAdapter, module_member, parse_qs, user_agent
from ..exceptions import AuthFailed


class BaseAuth(object):
    """A authentication backend that authenticates the user based on
    the provider response"""
    name = ''  # provider name, it's stored in database
    supports_inactive_user = False  # Django auth
    ID_KEY = None
    EXTRA_DATA = None
    GET_ALL_EXTRA_DATA = False
    REQUIRES_EMAIL_VALIDATION = False
    SEND_USER_AGENT = False
    SSL_PROTOCOL = None

    def __init__(self, strategy, redirect_uri=None):
        self.strategy = strategy
        self.redirect_uri = redirect_uri
        self.data = self.strategy.request_data()
        self.redirect_uri = self.strategy.absolute_uri(
            self.redirect_uri
        )

    def setting(self, name, default=None):
        """Return setting value from strategy"""
        return self.strategy.setting(name, default=default, backend=self)

    def start(self):
        if self.uses_redirect():
            return self.strategy.redirect(self.auth_url())
        else:
            return self.strategy.html(self.auth_html())

    def complete(self, *args, **kwargs):
        return self.auth_complete(*args, **kwargs)

    def auth_url(self):
        """Must return redirect URL to auth provider"""
        raise NotImplementedError('Implement in subclass')

    def auth_html(self):
        """Must return login HTML content returned by provider"""
        raise NotImplementedError('Implement in subclass')

    def auth_complete(self, *args, **kwargs):
        """Completes loging process, must return user instance"""
        raise NotImplementedError('Implement in subclass')

    def process_error(self, data):
        """Process data for errors, raise exception if needed.
        Call this method on any override of auth_complete."""
        pass

    def authenticate(self, *args, **kwargs):
        """Authenticate user using social credentials

        Authentication is made if this is the correct backend, backend
        verification is made by kwargs inspection for current backend
        name presence.
        """
        # Validate backend and arguments. Require that the Social Auth
        # response be passed in as a keyword argument, to make sure we
        # don't match the username/password calling conventions of
        # authenticate.
        if 'backend' not in kwargs or kwargs['backend'].name != self.name or \
           'strategy' not in kwargs or 'response' not in kwargs:
            return None

        self.strategy = kwargs.get('strategy') or self.strategy
        self.redirect_uri = kwargs.get('redirect_uri') or self.redirect_uri
        self.data = self.strategy.request_data()
        kwargs.setdefault('is_new', False)
        pipeline = self.strategy.get_pipeline(self)
        args, kwargs = self.strategy.clean_authenticate_args(*args, **kwargs)
        return self.pipeline(pipeline, *args, **kwargs)

    def pipeline(self, pipeline, pipeline_index=0, *args, **kwargs):
        out = self.run_pipeline(pipeline, pipeline_index, *args, **kwargs)
        if not isinstance(out, dict):
            return out
        user = out.get('user')
        if user:
            user.social_user = out.get('social')
            user.is_new = out.get('is_new')
        return user

    def disconnect(self, *args, **kwargs):
        pipeline = self.strategy.get_disconnect_pipeline(self)
        kwargs['name'] = self.name
        kwargs['user_storage'] = self.strategy.storage.user
        return self.run_pipeline(pipeline, *args, **kwargs)

    def run_pipeline(self, pipeline, pipeline_index=0, *args, **kwargs):
        out = kwargs.copy()
        out.setdefault('strategy', self.strategy)
        out.setdefault('backend', out.pop(self.name, None) or self)
        out.setdefault('request', self.strategy.request_data())
        out.setdefault('details', {})

        if not isinstance(pipeline_index, int) or \
           pipeline_index < 0 or \
           pipeline_index >= len(pipeline):
            pipeline_index = 0

        for idx, name in enumerate(pipeline[pipeline_index:]):
            out['pipeline_index'] = pipeline_index + idx
            func = module_member(name)
            result = func(*args, **out) or {}
            if not isinstance(result, dict):
                return result
            out.update(result)
        return out

    def extra_data(self, user, uid, response, details=None, *args, **kwargs):
        """Return default extra data to store in extra_data field"""
        data = {
            # store the last time authentication toke place
            'auth_time': int(time.time())
        }
        extra_data_entries = []
        if self.GET_ALL_EXTRA_DATA or self.setting('GET_ALL_EXTRA_DATA', False):
            extra_data_entries = response.keys()
        else:
            extra_data_entries = (self.EXTRA_DATA or []) + self.setting('EXTRA_DATA', [])
        for entry in extra_data_entries:
            if not isinstance(entry, (list, tuple)):
                entry = (entry,)
            size = len(entry)
            if size >= 1 and size <= 3:
                if size == 3:
                    name, alias, discard = entry
                elif size == 2:
                    (name, alias), discard = entry, False
                elif size == 1:
                    name = alias = entry[0]
                    discard = False
                value = response.get(name) or details.get(name) or details.get(alias)
                if discard and not value:
                    continue
                data[alias] = value
        return data

    def auth_allowed(self, response, details):
        """Return True if the user should be allowed to authenticate, by
        default check if email is whitelisted (if there's a whitelist)"""
        emails = self.setting('WHITELISTED_EMAILS', [])
        domains = self.setting('WHITELISTED_DOMAINS', [])
        email = details.get('email')
        allowed = True
        if email and (emails or domains):
            domain = email.split('@', 1)[1]
            allowed = email in emails or domain in domains
        return allowed

    def get_user_id(self, details, response):
        """Return a unique ID for the current user, by default from server
        response."""
        return response.get(self.ID_KEY)

    def get_user_details(self, response):
        """Must return user details in a know internal struct:
            {'username': <username if any>,
             'email': <user email if any>,
             'fullname': <user full name if any>,
             'first_name': <user first name if any>,
             'last_name': <user last name if any>}
        """
        raise NotImplementedError('Implement in subclass')

    def get_user_names(self, fullname='', first_name='', last_name=''):
        # Avoid None values
        fullname = fullname or ''
        first_name = first_name or ''
        last_name = last_name or ''
        if fullname and not (first_name or last_name):
            try:
                first_name, last_name = fullname.split(' ', 1)
            except ValueError:
                first_name = first_name or fullname or ''
                last_name = last_name or ''
        fullname = fullname or ' '.join((first_name, last_name))
        return fullname.strip(), first_name.strip(), last_name.strip()

    def get_user(self, user_id):
        """
        Return user with given ID from the User model used by this backend.
        This is called by django.contrib.auth.middleware.
        """
        return self.strategy.get_user(user_id)

    def continue_pipeline(self, partial):
        """Continue previous halted pipeline"""
        return self.strategy.authenticate(self,
                                          pipeline_index=partial.next_step,
                                          *partial.args,
                                          **partial.kwargs)

    def auth_extra_arguments(self):
        """Return extra arguments needed on auth process. The defaults can be
        overridden by GET parameters."""
        extra_arguments = self.setting('AUTH_EXTRA_ARGUMENTS', {}).copy()
        extra_arguments.update((key, self.data[key]) for key in extra_arguments
                                    if key in self.data)
        return extra_arguments

    def uses_redirect(self):
        """Return True if this provider uses redirect url method,
        otherwise return false."""
        return True

    def request(self, url, method='GET', *args, **kwargs):
        kwargs.setdefault('headers', {})
        if self.setting('VERIFY_SSL') is not None:
            kwargs.setdefault('verify', self.setting('VERIFY_SSL'))
        kwargs.setdefault('timeout', self.setting('REQUESTS_TIMEOUT') or
                                     self.setting('URLOPEN_TIMEOUT'))
        if self.SEND_USER_AGENT and 'User-Agent' not in kwargs['headers']:
            kwargs['headers']['User-Agent'] = self.setting('USER_AGENT') or \
                                              user_agent()

        try:
            if self.SSL_PROTOCOL:
                session = SSLHttpAdapter.ssl_adapter_session(self.SSL_PROTOCOL)
                response = session.request(method, url, *args, **kwargs)
            else:
                response = request(method, url, *args, **kwargs)
        except ConnectionError as err:
            raise AuthFailed(self, str(err))
        response.raise_for_status()
        return response

    def get_json(self, url, *args, **kwargs):
        return self.request(url, *args, **kwargs).json()

    def get_querystring(self, url, *args, **kwargs):
        return parse_qs(self.request(url, *args, **kwargs).text)

    def get_key_and_secret(self):
        """Return tuple with Consumer Key and Consumer Secret for current
        service provider. Must return (key, secret), order *must* be respected.
        """
        return self.setting('KEY'), self.setting('SECRET')



# -*- coding: utf-8 -*-
import sys
import re
import os

from os.path import join, dirname

from setuptools import setup


VERSION_RE = re.compile(r'__version__ = \'([\d\.]+)\'')

LONG_DESCRIPTION = """
Python Social Auth is an easy to setup social authentication/registration
mechanism with support for several frameworks and auth providers.

It implements a common interface to define new authentication
providers from third parties. And to bring support for more frameworks
and ORMs.
"""


def long_description():
    try:
        return open(join(dirname(__file__), 'README.md')).read()
    except IOError:
        return None


def read_version():
    with open('social_core/__init__.py') as file:
        version_line = [line for line in file.readlines()
                        if line.startswith('__version__')][0]
        return VERSION_RE.match(version_line).groups()[0]


def read_requirements(filename):
    with open(filename, 'r') as file:
        return [line for line in file.readlines() if not line.startswith('-')]


def read_tests_requirements(filename):
    return read_requirements('social_core/tests/{0}'.format(filename))


PY = os.environ.get("BUILD_VERSION") or sys.version_info[0]
requirements = read_requirements('requirements-base.txt')
# May be able to just use environment markers in requirements-base.txt
# at least on  setuptools 36.2.0 and up.
requirements_py2 = read_requirements('requirements-python2.txt')
requirements_py3 = read_requirements('requirements-python3.txt')
requirements_openidconnect = read_requirements('requirements-openidconnect.txt')
requirements_saml = read_requirements('requirements-saml-python%s.txt' % PY)
requirements_azuread = read_requirements('requirements-azuread.txt')

tests_requirements_base = read_tests_requirements('requirements-base.txt')
tests_requirements = tests_requirements_base + \
    read_tests_requirements('requirements-python%s.txt' % PY)

requirements_all = requirements_openidconnect + \
                   requirements_saml + \
                   requirements_azuread

tests_requirements = tests_requirements + requirements_all

setup(
    name='social-auth-core',
    version=read_version(),
    author='Matias Aguirre',
    author_email='matiasaguirre@gmail.com',
    description='Python social authentication made simple.',
    license='BSD',
    keywords='openid, oauth, saml, social auth',
    url='https://github.com/python-social-auth/social-core',
    packages=[
        'social_core',
        'social_core.backends',
        'social_core.pipeline',
        'social_core.tests',
        'social_core.tests.actions',
        'social_core.tests.backends',
        'social_core.tests.backends.data'
    ],
    long_description=long_description() or LONG_DESCRIPTION,
    install_requires=requirements,
    extras_require={
        'openidconnect': [requirements_openidconnect],
        'saml': [requirements_saml],
        'azuread': [requirements_azuread],
        'all': [requirements_all],
        ':python_version < "3.0"': [requirements_py2],
        ':python_version >= "3.0"': [requirements_py3],
    },
    classifiers=[
        'Development Status :: 4 - Beta',
        'Topic :: Internet',
        'License :: OSI Approved :: BSD License',
        'Intended Audience :: Developers',
        'Environment :: Web Environment',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3'
    ],
    package_data={
        'social_core/tests': [
            'social_core/tests/*.txt',
            'social_core/tests/testkey.pem'
        ]
    },
    include_package_data=True,
    tests_require=tests_requirements,
    test_suite='social_core.tests',
    zip_safe=False
)



"""Models mixins for Social Auth"""
import re
import time
import base64
import uuid
import warnings

from datetime import datetime, timedelta

import six

from openid.association import Association as OpenIdAssociation

from .exceptions import MissingBackend
from .backends.utils import get_backend


NO_ASCII_REGEX = re.compile(r'[^\x00-\x7F]+')
NO_SPECIAL_REGEX = re.compile(r'[^\w.@+_-]+', re.UNICODE)


class UserMixin(object):
    user = ''
    provider = ''
    uid = None
    extra_data = None

    def get_backend(self, strategy):
        return get_backend(strategy.get_backends(), self.provider)

    def get_backend_instance(self, strategy):
        try:
            backend_class = self.get_backend(strategy)
        except MissingBackend:
            return None
        else:
            return backend_class(strategy=strategy)

    @property
    def access_token(self):
        """Return access_token stored in extra_data or None"""
        return self.extra_data.get('access_token')

    @property
    def tokens(self):
        warnings.warn('tokens is deprecated, use access_token instead')
        return self.access_token

    def refresh_token(self, strategy, *args, **kwargs):
        token = self.extra_data.get('refresh_token') or \
                self.extra_data.get('access_token')
        backend = self.get_backend(strategy)
        if token and backend and hasattr(backend, 'refresh_token'):
            backend = backend(strategy=strategy)
            response = backend.refresh_token(token, *args, **kwargs)
            extra_data = backend.extra_data(self,
                                            self.uid,
                                            response,
                                            self.extra_data)
            if self.set_extra_data(extra_data):
                self.save()

    def expiration_timedelta(self):
        """Return provider session live seconds. Returns a timedelta ready to
        use with session.set_expiry().

        If provider returns a timestamp instead of session seconds to live, the
        timedelta is inferred from current time (using UTC timezone). None is
        returned if there's no value stored or it's invalid.
        """
        if self.extra_data and 'expires' in self.extra_data:
            try:
                expires = int(self.extra_data.get('expires'))
            except (ValueError, TypeError):
                return None

            now = datetime.utcnow()

            # Detect if expires is a timestamp
            if expires > time.mktime(now.timetuple()):
                # expires is a datetime, return the remaining difference
                return datetime.utcfromtimestamp(expires) - now
            else:
                # expires is the time to live seconds since creation,
                # check against auth_time if present, otherwise return
                # the value
                auth_time = self.extra_data.get('auth_time')
                if auth_time:
                    reference = datetime.utcfromtimestamp(auth_time)
                    return (reference + timedelta(seconds=expires)) - now
                else:
                    return timedelta(seconds=expires)

    def expiration_datetime(self):
        # backward compatible alias
        return self.expiration_timedelta()

    def access_token_expired(self):
        expiration = self.expiration_timedelta()
        return expiration and expiration.total_seconds() <= 0

    def get_access_token(self, strategy):
        """Returns a valid access token."""
        if self.access_token_expired():
            self.refresh_token(strategy)
        return self.access_token

    def set_extra_data(self, extra_data=None):
        if extra_data and self.extra_data != extra_data:
            if self.extra_data and not isinstance(
                    self.extra_data, six.string_types):
                self.extra_data.update(extra_data)
            else:
                self.extra_data = extra_data
            return True

    @classmethod
    def clean_username(cls, value):
        """Clean username removing any unsupported character"""
        value = NO_ASCII_REGEX.sub('', value)
        value = NO_SPECIAL_REGEX.sub('', value)
        return value

    @classmethod
    def changed(cls, user):
        """The given user instance is ready to be saved"""
        raise NotImplementedError('Implement in subclass')

    @classmethod
    def get_username(cls, user):
        """Return the username for given user"""
        raise NotImplementedError('Implement in subclass')

    @classmethod
    def user_model(cls):
        """Return the user model"""
        raise NotImplementedError('Implement in subclass')

    @classmethod
    def username_max_length(cls):
        """Return the max length for username"""
        raise NotImplementedError('Implement in subclass')

    @classmethod
    def allowed_to_disconnect(cls, user, backend_name, association_id=None):
        """Return if it's safe to disconnect the social account for the
        given user"""
        raise NotImplementedError('Implement in subclass')

    @classmethod
    def disconnect(cls, entry):
        """Disconnect the social account for the given user"""
        raise NotImplementedError('Implement in subclass')

    @classmethod
    def user_exists(cls, *args, **kwargs):
        """
        Return True/False if a User instance exists with the given arguments.
        Arguments are directly passed to filter() manager method.
        """
        raise NotImplementedError('Implement in subclass')

    @classmethod
    def create_user(cls, *args, **kwargs):
        """Create a user instance"""
        raise NotImplementedError('Implement in subclass')

    @classmethod
    def get_user(cls, pk):
        """Return user instance for given id"""
        raise NotImplementedError('Implement in subclass')

    @classmethod
    def get_users_by_email(cls, email):
        """Return users instances for given email address"""
        raise NotImplementedError('Implement in subclass')

    @classmethod
    def get_social_auth(cls, provider, uid):
        """Return UserSocialAuth for given provider and uid"""
        raise NotImplementedError('Implement in subclass')

    @classmethod
    def get_social_auth_for_user(cls, user, provider=None, id=None):
        """Return all the UserSocialAuth instances for given user"""
        raise NotImplementedError('Implement in subclass')

    @classmethod
    def create_social_auth(cls, user, uid, provider):
        """Create a UserSocialAuth instance for given user"""
        raise NotImplementedError('Implement in subclass')


class NonceMixin(object):
    """One use numbers"""
    server_url = ''
    timestamp = 0
    salt = ''

    @classmethod
    def use(cls, server_url, timestamp, salt):
        """Create a Nonce instance"""
        raise NotImplementedError('Implement in subclass')


class AssociationMixin(object):
    """OpenId account association"""
    server_url = ''
    handle = ''
    secret = ''
    issued = 0
    lifetime = 0
    assoc_type = ''

    @classmethod
    def oids(cls, server_url, handle=None):
        kwargs = {'server_url': server_url}
        if handle is not None:
            kwargs['handle'] = handle
        return sorted([
            (assoc.id, cls.openid_association(assoc))
            for assoc in cls.get(**kwargs)
        ], key=lambda x: x[1].issued, reverse=True)

    @classmethod
    def openid_association(cls, assoc):
        secret = assoc.secret
        if not isinstance(secret, six.binary_type):
            secret = secret.encode()
        return OpenIdAssociation(assoc.handle, base64.decodestring(secret),
                                 assoc.issued, assoc.lifetime,
                                 assoc.assoc_type)

    @classmethod
    def store(cls, server_url, association):
        """Create an Association instance"""
        raise NotImplementedError('Implement in subclass')

    @classmethod
    def get(cls, *args, **kwargs):
        """Get an Association instance"""
        raise NotImplementedError('Implement in subclass')

    @classmethod
    def remove(cls, ids_to_delete):
        """Remove an Association instance"""
        raise NotImplementedError('Implement in subclass')


class CodeMixin(object):
    email = ''
    code = ''
    verified = False

    def verify(self):
        self.verified = True
        self.save()

    @classmethod
    def generate_code(cls):
        return uuid.uuid4().hex

    @classmethod
    def make_code(cls, email):
        code = cls()
        code.email = email
        code.code = cls.generate_code()
        code.verified = False
        code.save()
        return code

    @classmethod
    def get_code(cls, code):
        raise NotImplementedError('Implement in subclass')


class PartialMixin(object):
    token = ''
    data = ''
    next_step = ''
    backend = ''

    @property
    def args(self):
        return self.data.get('args', [])

    @args.setter
    def args(self, value):
        self.data['args'] = value

    @property
    def kwargs(self):
        return self.data.get('kwargs', {})

    @kwargs.setter
    def kwargs(self, value):
        self.data['kwargs'] = value

    def extend_kwargs(self, values):
        self.data['kwargs'].update(values)

    @classmethod
    def generate_token(cls):
        return uuid.uuid4().hex

    @classmethod
    def load(cls, token):
        raise NotImplementedError('Implement in subclass')

    @classmethod
    def destroy(cls, token):
        raise NotImplementedError('Implement in subclass')

    @classmethod
    def prepare(cls, backend, next_step, data):
        partial = cls()
        partial.backend = backend
        partial.next_step = next_step
        partial.data = data
        partial.token = cls.generate_token()
        return partial

    @classmethod
    def store(cls, partial):
        partial.save()
        return partial


class BaseStorage(object):
    user = UserMixin
    nonce = NonceMixin
    association = AssociationMixin
    code = CodeMixin
    partial = PartialMixin

    @classmethod
    def is_integrity_error(cls, exception):
        """Check if given exception flags an integrity error in the DB"""
        raise NotImplementedError('Implement in subclass')



import six

from requests_oauthlib import OAuth1
from oauthlib.oauth1 import SIGNATURE_TYPE_AUTH_HEADER

from six.moves.urllib_parse import urlencode, unquote

from ..utils import url_add_parameters, parse_qs, handle_http_errors, \
                    constant_time_compare
from ..exceptions import AuthFailed, AuthCanceled, AuthUnknownError, \
                         AuthMissingParameter, AuthStateMissing, \
                         AuthStateForbidden, AuthTokenError
from .base import BaseAuth


class OAuthAuth(BaseAuth):
    """OAuth authentication backend base class.

    Also settings will be inspected to get more values names that should be
    stored on extra_data field. Setting name is created from current backend
    name (all uppercase) plus _EXTRA_DATA.

    access_token is always stored.

    URLs settings:
        AUTHORIZATION_URL       Authorization service url
        ACCESS_TOKEN_URL        Access token URL
    """
    AUTHORIZATION_URL = ''
    ACCESS_TOKEN_URL = ''
    ACCESS_TOKEN_METHOD = 'GET'
    REVOKE_TOKEN_URL = None
    REVOKE_TOKEN_METHOD = 'POST'
    ID_KEY = 'id'
    SCOPE_PARAMETER_NAME = 'scope'
    DEFAULT_SCOPE = None
    SCOPE_SEPARATOR = ' '
    REDIRECT_STATE = False
    STATE_PARAMETER = False

    def extra_data(self, user, uid, response, details=None, *args, **kwargs):
        """Return access_token and extra defined names to store in
        extra_data field"""
        data = super(OAuthAuth, self).extra_data(user, uid, response, details,
                                                 *args, **kwargs)
        data['access_token'] = response.get('access_token', '') or \
                               kwargs.get('access_token')
        return data

    def state_token(self):
        """Generate csrf token to include as state parameter."""
        return self.strategy.random_string(32)

    def get_or_create_state(self):
        if self.STATE_PARAMETER or self.REDIRECT_STATE:
            # Store state in session for further request validation. The state
            # value is passed as state parameter (as specified in OAuth2 spec),
            # but also added to redirect, that way we can still verify the
            # request if the provider doesn't implement the state parameter.
            # Reuse token if any.
            name = self.name + '_state'
            state = self.strategy.session_get(name)
            if state is None:
                state = self.state_token()
                self.strategy.session_set(name, state)
        else:
            state = None
        return state

    def get_session_state(self):
        return self.strategy.session_get(self.name + '_state')

    def get_request_state(self):
        request_state = self.data.get('state') or \
                        self.data.get('redirect_state')
        if request_state and isinstance(request_state, list):
            request_state = request_state[0]
        return request_state

    def validate_state(self):
        """Validate state value. Raises exception on error, returns state
        value if valid."""
        if not self.STATE_PARAMETER and not self.REDIRECT_STATE:
            return None
        state = self.get_session_state()
        request_state = self.get_request_state()
        if not request_state:
            raise AuthMissingParameter(self, 'state')
        elif not state:
            raise AuthStateMissing(self, 'state')
        elif not constant_time_compare(request_state, state):
            raise AuthStateForbidden(self)
        else:
            return state

    def get_redirect_uri(self, state=None):
        """Build redirect with redirect_state parameter."""
        uri = self.redirect_uri
        if self.REDIRECT_STATE and state:
            uri = url_add_parameters(uri, {'redirect_state': state})
        return uri

    def get_scope(self):
        """Return list with needed access scope"""
        scope = self.setting('SCOPE', [])
        if not self.setting('IGNORE_DEFAULT_SCOPE', False):
            scope = scope + (self.DEFAULT_SCOPE or [])
        return scope

    def get_scope_argument(self):
        param = {}
        scope = self.get_scope()
        if scope:
            param[self.SCOPE_PARAMETER_NAME] = self.SCOPE_SEPARATOR.join(scope)
        return param

    def user_data(self, access_token, *args, **kwargs):
        """Loads user data from service. Implement in subclass"""
        return {}

    def authorization_url(self):
        return self.AUTHORIZATION_URL

    def access_token_url(self):
        return self.ACCESS_TOKEN_URL

    def revoke_token_url(self, token, uid):
        return self.REVOKE_TOKEN_URL

    def revoke_token_params(self, token, uid):
        return {}

    def revoke_token_headers(self, token, uid):
        return {}

    def process_revoke_token_response(self, response):
        return response.status_code == 200

    def revoke_token(self, token, uid):
        if self.REVOKE_TOKEN_URL:
            url = self.revoke_token_url(token, uid)
            params = self.revoke_token_params(token, uid)
            headers = self.revoke_token_headers(token, uid)
            data = urlencode(params) if self.REVOKE_TOKEN_METHOD != 'GET' \
                                     else None
            response = self.request(url, params=params, headers=headers,
                                    data=data, method=self.REVOKE_TOKEN_METHOD)
            return self.process_revoke_token_response(response)


class BaseOAuth1(OAuthAuth):
    """Consumer based mechanism OAuth authentication, fill the needed
    parameters to communicate properly with authentication service.

    URLs settings:
        REQUEST_TOKEN_URL       Request token URL

    """
    REQUEST_TOKEN_URL = ''
    REQUEST_TOKEN_METHOD = 'GET'
    OAUTH_TOKEN_PARAMETER_NAME = 'oauth_token'
    REDIRECT_URI_PARAMETER_NAME = 'redirect_uri'
    UNATHORIZED_TOKEN_SUFIX = 'unauthorized_token_name'

    def auth_url(self):
        """Return redirect url"""
        token = self.set_unauthorized_token()
        return self.oauth_authorization_request(token)

    def process_error(self, data):
        if 'oauth_problem' in data:
            if data['oauth_problem'] == 'user_refused':
                raise AuthCanceled(self, 'User refused the access')
            raise AuthUnknownError(self, 'Error was ' + data['oauth_problem'])

    @handle_http_errors
    def auth_complete(self, *args, **kwargs):
        """Return user, might be logged in"""
        # Multiple unauthorized tokens are supported (see #521)
        self.process_error(self.data)
        self.validate_state()
        token = self.get_unauthorized_token()
        access_token = self.access_token(token)
        return self.do_auth(access_token, *args, **kwargs)

    @handle_http_errors
    def do_auth(self, access_token, *args, **kwargs):
        """Finish the auth process once the access_token was retrieved"""
        if not isinstance(access_token, dict):
            access_token = parse_qs(access_token)
        data = self.user_data(access_token)
        if data is not None and 'access_token' not in data:
            data['access_token'] = access_token
        kwargs.update({'response': data, 'backend': self})
        return self.strategy.authenticate(*args, **kwargs)

    def get_unauthorized_token(self):
        name = self.name + self.UNATHORIZED_TOKEN_SUFIX
        unauthed_tokens = self.strategy.session_get(name, [])
        if not unauthed_tokens:
            raise AuthTokenError(self, 'Missing unauthorized token')

        data_token = self.data.get(self.OAUTH_TOKEN_PARAMETER_NAME)

        if data_token is None:
            raise AuthTokenError(self, 'Missing unauthorized token')

        token = None
        for utoken in unauthed_tokens:
            orig_utoken = utoken
            if not isinstance(utoken, dict):
                utoken = parse_qs(utoken)
            if utoken.get(self.OAUTH_TOKEN_PARAMETER_NAME) == data_token:
                self.strategy.session_set(name, list(set(unauthed_tokens) -
                                                     set([orig_utoken])))
                token = utoken
                break
        else:
            raise AuthTokenError(self, 'Incorrect tokens')
        return token

    def set_unauthorized_token(self):
        token = self.unauthorized_token()
        name = self.name + self.UNATHORIZED_TOKEN_SUFIX
        tokens = self.strategy.session_get(name, []) + [token]
        self.strategy.session_set(name, tokens)
        return token

    def request_token_extra_arguments(self):
        """Return extra arguments needed on request-token process"""
        return self.setting('REQUEST_TOKEN_EXTRA_ARGUMENTS', {})

    def unauthorized_token(self):
        """Return request for unauthorized token (first stage)"""
        params = self.request_token_extra_arguments()
        params.update(self.get_scope_argument())
        key, secret = self.get_key_and_secret()
        # decoding='utf-8' produces errors with python-requests on Python3
        # since the final URL will be of type bytes
        decoding = None if six.PY3 else 'utf-8'
        state = self.get_or_create_state()
        response = self.request(
            self.REQUEST_TOKEN_URL,
            params=params,
            auth=OAuth1(key, secret, callback_uri=self.get_redirect_uri(state),
                        decoding=decoding),
            method=self.REQUEST_TOKEN_METHOD
        )
        content = response.content
        if response.encoding or response.apparent_encoding:
            content = content.decode(response.encoding or
                                     response.apparent_encoding)
        else:
            content = response.content.decode()
        return content

    def oauth_authorization_request(self, token):
        """Generate OAuth request to authorize token."""
        if not isinstance(token, dict):
            token = parse_qs(token)
        params = self.auth_extra_arguments() or {}
        params.update(self.get_scope_argument())
        params[self.OAUTH_TOKEN_PARAMETER_NAME] = token.get(
            self.OAUTH_TOKEN_PARAMETER_NAME
        )
        state = self.get_or_create_state()
        params[self.REDIRECT_URI_PARAMETER_NAME] = self.get_redirect_uri(state)
        return '{0}?{1}'.format(self.authorization_url(), urlencode(params))

    def oauth_auth(self, token=None, oauth_verifier=None,
                   signature_type=SIGNATURE_TYPE_AUTH_HEADER):
        key, secret = self.get_key_and_secret()
        oauth_verifier = oauth_verifier or self.data.get('oauth_verifier')
        if token:
            resource_owner_key = token.get('oauth_token')
            resource_owner_secret = token.get('oauth_token_secret')
            if not resource_owner_key:
                raise AuthTokenError(self, 'Missing oauth_token')
            if not resource_owner_secret:
                raise AuthTokenError(self, 'Missing oauth_token_secret')
        else:
            resource_owner_key = None
            resource_owner_secret = None
        # decoding='utf-8' produces errors with python-requests on Python3
        # since the final URL will be of type bytes
        decoding = None if six.PY3 else 'utf-8'
        state = self.get_or_create_state()
        return OAuth1(key, secret,
                      resource_owner_key=resource_owner_key,
                      resource_owner_secret=resource_owner_secret,
                      callback_uri=self.get_redirect_uri(state),
                      verifier=oauth_verifier,
                      signature_type=signature_type,
                      decoding=decoding)

    def oauth_request(self, token, url, params=None, method='GET'):
        """Generate OAuth request, setups callback url"""
        return self.request(url, method=method, params=params,
                            auth=self.oauth_auth(token))

    def access_token(self, token):
        """Return request for access token value"""
        return self.get_querystring(self.access_token_url(),
                                    auth=self.oauth_auth(token),
                                    method=self.ACCESS_TOKEN_METHOD)


class BaseOAuth2(OAuthAuth):
    """Base class for OAuth2 providers.

    OAuth2 draft details at:
        http://tools.ietf.org/html/draft-ietf-oauth-v2-10
    """
    REFRESH_TOKEN_URL = None
    REFRESH_TOKEN_METHOD = 'POST'
    RESPONSE_TYPE = 'code'
    REDIRECT_STATE = True
    STATE_PARAMETER = True

    def auth_params(self, state=None):
        client_id, client_secret = self.get_key_and_secret()
        params = {
            'client_id': client_id,
            'redirect_uri': self.get_redirect_uri(state)
        }
        if self.STATE_PARAMETER and state:
            params['state'] = state
        if self.RESPONSE_TYPE:
            params['response_type'] = self.RESPONSE_TYPE
        return params

    def auth_url(self):
        """Return redirect url"""
        state = self.get_or_create_state()
        params = self.auth_params(state)
        params.update(self.get_scope_argument())
        params.update(self.auth_extra_arguments())
        params = urlencode(params)
        if not self.REDIRECT_STATE:
            # redirect_uri matching is strictly enforced, so match the
            # providers value exactly.
            params = unquote(params)
        return '{0}?{1}'.format(self.authorization_url(), params)

    def auth_complete_params(self, state=None):
        client_id, client_secret = self.get_key_and_secret()
        return {
            'grant_type': 'authorization_code',  # request auth code
            'code': self.data.get('code', ''),  # server response code
            'client_id': client_id,
            'client_secret': client_secret,
            'redirect_uri': self.get_redirect_uri(state)
        }

    def auth_complete_credentials(self):
        return None

    def auth_headers(self):
        return {'Content-Type': 'application/x-www-form-urlencoded',
                'Accept': 'application/json'}

    def extra_data(self, user, uid, response, details=None, *args, **kwargs):
        """Return access_token, token_type, and extra defined names to store in
            extra_data field"""
        data = super(BaseOAuth2, self).extra_data(user, uid, response,
                                                  details=details,
                                                  *args, **kwargs)
        data['token_type'] = response.get('token_type') or \
                             kwargs.get('token_type')
        return data

    def request_access_token(self, *args, **kwargs):
        return self.get_json(*args, **kwargs)

    def process_error(self, data):
        if data.get('error'):
            if data['error'] == 'denied' or data['error'] == 'access_denied':
                raise AuthCanceled(self, data.get('error_description', ''))
            raise AuthFailed(self, data.get('error_description') or
                                   data['error'])
        elif 'denied' in data:
            raise AuthCanceled(self, data['denied'])

    @handle_http_errors
    def auth_complete(self, *args, **kwargs):
        """Completes login process, must return user instance"""
        self.process_error(self.data)
        state = self.validate_state()
        data, params = None, None
        if self.ACCESS_TOKEN_METHOD == 'GET':
            params = self.auth_complete_params(state)
        else:
            data = self.auth_complete_params(state)

        response = self.request_access_token(
            self.access_token_url(),
            data=data,
            params=params,
            headers=self.auth_headers(),
            auth=self.auth_complete_credentials(),
            method=self.ACCESS_TOKEN_METHOD
        )
        self.process_error(response)
        return self.do_auth(response['access_token'], response=response,
                            *args, **kwargs)

    @handle_http_errors
    def do_auth(self, access_token, *args, **kwargs):
        """Finish the auth process once the access_token was retrieved"""
        data = self.user_data(access_token, *args, **kwargs)
        response = kwargs.get('response') or {}
        response.update(data or {})
        if 'access_token' not in response:
            response['access_token'] = access_token
        kwargs.update({'response': response, 'backend': self})
        return self.strategy.authenticate(*args, **kwargs)

    def refresh_token_params(self, token, *args, **kwargs):
        client_id, client_secret = self.get_key_and_secret()
        return {
            'refresh_token': token,
            'grant_type': 'refresh_token',
            'client_id': client_id,
            'client_secret': client_secret
        }

    def process_refresh_token_response(self, response, *args, **kwargs):
        return response.json()

    def refresh_token(self, token, *args, **kwargs):
        params = self.refresh_token_params(token, *args, **kwargs)
        url = self.refresh_token_url()
        method = self.REFRESH_TOKEN_METHOD
        key = 'params' if method == 'GET' else 'data'
        request_args = {'headers': self.auth_headers(),
                        'method': method,
                        key: params}
        request = self.request(url, **request_args)
        return self.process_refresh_token_response(request, *args, **kwargs)

    def refresh_token_url(self):
        return self.REFRESH_TOKEN_URL or self.access_token_url()



from uuid import uuid4

from ..utils import slugify, module_member


USER_FIELDS = ['username', 'email']


def get_username(strategy, details, backend, user=None, *args, **kwargs):
    if 'username' not in backend.setting('USER_FIELDS', USER_FIELDS):
        return
    storage = strategy.storage

    if not user:
        email_as_username = strategy.setting('USERNAME_IS_FULL_EMAIL', False)
        uuid_length = strategy.setting('UUID_LENGTH', 16)
        max_length = storage.user.username_max_length()
        do_slugify = strategy.setting('SLUGIFY_USERNAMES', False)
        do_clean = strategy.setting('CLEAN_USERNAMES', True)

        if do_clean:
            override_clean = strategy.setting('CLEAN_USERNAME_FUNCTION')
            if override_clean:
                clean_func = module_member(override_clean)
            else:
                clean_func = storage.user.clean_username
        else:
            clean_func = lambda val: val

        if do_slugify:
            override_slug = strategy.setting('SLUGIFY_FUNCTION')
            if override_slug:
                slug_func = module_member(override_slug)
            else:
                slug_func = slugify
        else:
            slug_func = lambda val: val

        if email_as_username and details.get('email'):
            username = details['email']
        elif details.get('username'):
            username = details['username']
        else:
            username = uuid4().hex

        short_username = (username[:max_length - uuid_length]
                          if max_length is not None
                          else username)
        final_username = slug_func(clean_func(username[:max_length]))

        # Generate a unique username for current user using username
        # as base but adding a unique hash at the end. Original
        # username is cut to avoid any field max_length.
        # The final_username may be empty and will skip the loop.
        while not final_username or \
              storage.user.user_exists(username=final_username):
            username = short_username + uuid4().hex[:uuid_length]
            final_username = slug_func(clean_func(username[:max_length]))
    else:
        final_username = storage.user.get_username(user)
    return {'username': final_username}


def create_user(strategy, details, backend, user=None, *args, **kwargs):
    if user:
        return {'is_new': False}

    fields = dict((name, kwargs.get(name, details.get(name)))
                  for name in backend.setting('USER_FIELDS', USER_FIELDS))
    if not fields:
        return

    return {
        'is_new': True,
        'user': strategy.create_user(**fields)
    }


def user_details(strategy, details, user=None, *args, **kwargs):
    """Update user details using data from provider."""
    if not user:
        return

    changed = False  # flag to track changes
    protected = ('username', 'id', 'pk', 'email') + \
                tuple(strategy.setting('PROTECTED_USER_FIELDS', []))

    # Update user model attributes with the new data sent by the current
    # provider. Update on some attributes is disabled by default, for
    # example username and id fields. It's also possible to disable update
    # on fields defined in SOCIAL_AUTH_PROTECTED_FIELDS.
    for name, value in details.items():
        if value is None or not hasattr(user, name) or name in protected:
            continue

        # Check https://github.com/omab/python-social-auth/issues/671
        current_value = getattr(user, name, None)
        if current_value or current_value == value:
            continue

        changed = True
        setattr(user, name, value)

    if changed:
        strategy.storage.user.changed(user)



import json

from .oauth import OAuth2Test


class CoinbaseOAuth2Test(OAuth2Test):
    backend_path = 'social_core.backends.coinbase.CoinbaseOAuth2'
    user_data_url = 'https://api.coinbase.com/v2/user'
    expected_username = 'satoshi_nakomoto'
    access_token_body = json.dumps({
        'access_token': 'foobar',
        'token_type': 'bearer'
    })
    user_data_body = json.dumps({
        "data": {
            "id": "1A1zP1eP5QGefi2DMPTfTL5SLmv7DivfNa",
            'name': "Satoshi Nakamoto",
            "username": "satoshi_nakomoto",
            "profile_location": None,
            "profile_bio": None,
            "profile_url": "https://coinbase.com/satoshi_nakomoto",
            "avatar_url": None,
            "resource": "user",
            "resource_path": "/v2/user"
          }
    })

    def test_login(self):
        self.do_login()

    def test_partial_pipeline(self):
        self.do_partial_pipeline()



"""
Bungie OAuth2 backend
"""
from social_core.backends.oauth import BaseOAuth2
from django.conf import settings


class BungieOAuth2(BaseOAuth2):

    name = 'bungie'
    ID_KEY = 'membership_id'
    AUTHORIZATION_URL = 'https://www.bungie.net/en/oauth/authorize/'
    ACCESS_TOKEN_URL = 'https://www.bungie.net/platform/app/oauth/token/'
    REFRESH_TOKEN_URL = 'https://www.bungie.net/platform/app/oauth/token/'
    ACCESS_TOKEN_METHOD = 'POST'
    REDIRECT_STATE = False
    EXTRA_DATA = [
        ('refresh_token', 'refresh_token', True),
        ('access_token', 'access_token', True),
        ('expires_in', 'expires'),
        ('membership_id', 'membership_id'),
        ('refresh_expires_in', 'refresh_expires_in')
    ]

    def auth_html(self):
        """Abstract Method Inclusion"""
        pass

    def auth_headers(self):
        """Adds X-API-KEY and Origin"""
        return {'X-API-KEY': settings.SOCIAL_AUTH_BUNGIE_API_KEY,
                'Content-Type': 'application/x-www-form-urlencoded',
                'Origin': settings.SOCIAL_AUTH_BUNGIE_ORIGIN,
                'Accept': 'application/json'
                }

    def make_bungie_request(self, url, access_token, kwargs):
        """Helper function to get username data keyed off displayName"""
        print('ENTERING MAKE BUNGIE REQUEST')
        headers = self.auth_headers()
        print(repr(headers))
        auth_header = {'Authorization': 'Bearer ' + access_token}
        headers.update(auth_header)
        import requests as python_requests
        r = python_requests.get(url, headers=headers)
        this_json = r.json()
        return this_json

    def auth_complete(self, *args, **kwargs):
        """Completes login process, must return user instance"""
        self.process_error(self.data)
        state = self.validate_state()
        response = self.request_access_token(
            self.access_token_url(),
            data=self.auth_complete_params(state),
            headers=self.auth_headers(),
            auth=self.auth_complete_credentials(),
            method=self.ACCESS_TOKEN_METHOD
        )
        self.process_error(response)
        return self.do_auth(response['access_token'], response=response, *args, **kwargs)

    def do_auth(self, access_token, *args, **kwargs):
        """Finish the auth process once the access_token was retrieved"""
        data = self.user_data(access_token, *args, **kwargs)
        response = kwargs.get('response') or {}
        response.update(data or {})
        if 'access_token' not in response:
            response['Response']['access_token']['value'] = access_token
        kwargs.update({'response': response, 'backend': self})
        return self.strategy.authenticate(*args, **kwargs)

    def user_data(self, access_token, *args, **kwargs):
        """Grab user profile information from Bunige"""
        membership_id = kwargs['response']['membership_id']
        url = 'https://www.bungie.net/Platform/User/GetBungieNetUser/'
        this_json = self.make_bungie_request(url, access_token, kwargs)
        username = this_json['Response']['user']['displayName']
        return {'username': username, 'uid': membership_id}

    def get_user_details(self, response, *args, **kwargs):
        """Return user details from Bungie account"""
        username = response['username']
        uid = response['uid']
        bnId = response['bnId']
        return {
            'first_name': username,
            'username': username,
            'uid': uid,
        }
